###########################
#Convergent Cross Mapping #
########################### 

#######################################################
#Contact: R. Huffaker                                 #   
#Professor                                            #
#Department of Agricultural and Biological Engineering#
#Institute of Food and Agricultural Sciences          #
#University of Florida                                #
#rhuffaker@ufl.edu                                    #
#######################################################

##################################################################################
#Description: This R-Code runs convergent cross mapping (ccm)for multiple        #
#SSA-filtered time series using R package 'multispatialCCM'. The purpose of ccm  # 
#is to identify causal networks defining reconstructed dynamics in weakly-coupled#
#dynamic systems. The logic is: If x2 drives x1, then x2's dynamics are encoded  #
#into the dynamics of x1, and the attracting manifold for x1 (Mx1) can be used   #
#to cross-estimate x2. If x2 and x1 have a bi-causal relationship, then each can #
#be cross mapped by the attracting manifold of the other. A choice of modified   #
#Nash-Sutcliffe model efficiency measures are used to determine predictive skill.#                               #
##################################################################################

##################################################
#Input: Spreadsheet whose columns are time series#
##################################################

####################################################################################
#Output: Matrices "ccm_corr_name(time series)" for each time series. For example,  # 
#the matrix labeled "ccm_corr_x" determines whether x drives other variables in the#
#data set. Whether x drives the kth variable is given by the vector of correlation #
#coefficients in the kth column. The code coverts this output into an adjacency    #
#matrix to use in social network plots.                                            #
####################################################################################

###############################################################################
#References:                                                                  #
#Package 'multispatialCCM', Author: Clark, Adam                               #
#Package 'tseriesChaos', Author: Antonio, Fabio Di Narzo                      #
#Sugihara et al. (2012). Detecting Causality in Complex Ecosystems.           # 
# Science 338:496-500.                                                        #
#Clark, A., Isbell, F., Deyle, E., Cowles, J., Tilman, G., Sugihara, G. (2015)#
# Ecology 96(5):1174-1181.                                                    #
###############################################################################

rm(list=ls(all=TRUE))

#############################
#Load User-Defined Functions#
#############################
dump("embed_udf", file="embed_udf.R")
source("embed_udf.R")

################
#Load Libraries#
################
library(tseriesChaos)
library(multispatialCCM)

###########
#Load Data#
###########
ts<-read.csv(".csv")
data<-ts[:]
data<-na.omit(data)
data<-as.matrix(data)
names<-colnames(data)

############
#Parameters#
############
#(1-frac) is the number of points used to calculate the asymptote of the
#ccm curve (plotting the correlation coefficient against the library of
#points on the attractor used to cross predict the other variables
frac<-0.66              

#The code runs through two loops.  Loop j specifies the forcing process (x) 
#to be tested against response process (y) specified by loop k.
n.vars<-ncol(data)
hold.j<-matrix(0,n.vars,n.vars) 
for(j in 1:n.vars){
  print("j");print(j)
  x<-data[,j]
  #embed data
  results.embed<-embed_udf(x)
  m<-results.embed[[3]]
  d<-results.embed[[1]]
  lib<-((d*(m-1)+(m+1)):length(x)-m+2)
  #if (m>3) {m=3}
  
  hold.xy<-matrix(0,length(lib),n.vars) #ccm correlations 
  for(k in 1:n.vars){
    print("k");print(k)
    y<-data[,k]
    ccm.xy<-CCM_boot(x,y,E=m,tau=d,DesiredL=lib,iterations=100)
    corr.xy<-ccm.xy$rho  #ccm correlation coefficients for each library
    if(length(corr.xy)<length(lib)){length(corr.xy)=length(lib)}
    hold.xy[,k]<-corr.xy  #columns are ccm coefficients
    
    } #end loop k
    corr.mat.xy<-hold.xy;corr.mat.xy<-na.omit(corr.mat.xy)
 
  #######################################
  #Export .csv file for each corr.mat.xy#
  ####################################### 
  #Reading files: For example, the matrix labeled "ccm_corr_x" determines whether 
  #x drives other variables in the data set. Whether x drives the kth variable is 
  #given by the vector of correlation coefficients in the kth column.
  setwd(dir.results)
  file<-as.character("ccm.csv")
  names.var<-as.matrix(colnames(data));
  file<- as.character(paste(substr(file,1,nchar(file)-4),
           "_corr_", names.var[j],".csv", sep=""))
  print(file)
  write.table(corr.mat.xy,file,col.names=colnames(data),
              row.names=FALSE,sep=",")

  ###########################################################################
  #Convert corr.mat.xy into weighted adjacency matrix for network plot.     #
  #Rule:  Average last fractional group of correlations. Divide corr.mat.xy #
  #into two submatrices with submatrix containing final rows equal to       #
  #frac.last*nrow(nse.mat).                                                 #
  ###########################################################################
  n<-frac*nrow(corr.mat.xy);n<-round(n)
  corr.mat.xy.1<-corr.mat.xy[(n+1):nrow(corr.mat.xy),]
  corr.mat.xy.ave<-colMeans(corr.mat.xy.1)
  hold.j[,j]<-corr.mat.xy.ave
} #end loop j
corr.mat.xy.ave<-hold.j
corr.mat.xy.ave[is.na(corr.mat.xy.ave)] <- 0  #set nas = 0
#Insert row and column labels in w.adj.scaled
row.names <- c(names.var)
col.names<-c("",names.var)
matrix1<-cbind(row.names,corr.mat.xy.ave)
matrix2<-rbind(col.names,matrix1)
write.table(matrix2,"corr.mat.xy.ave.csv",col.names=FALSE,row.names=FALSE,sep=",")

##########################################
#Compute scaled weighted adjacency matrix#
##########################################
mat.1<-corr.mat.xy.ave
for(row in 1:nrow(corr.mat.xy.ave)){#row
   for(col in 1:nrow(corr.mat.xy.ave)){#column
      if(0.9<=mat.1[row,col]){mat.1[row,col]=2}
      if(0.8<=mat.1[row,col]&mat.1[row,col]<0.9){mat.1[row,col]=1}
      #if(0.60<mat.1[row,col]&mat.1[row,col]<0.70){mat.1[row,col]=1}
      if(mat.1[row,col]<0.80){mat.1[row,col]=0}
   } #end col
} #end row
#Reading file: Driving variable listed in column. This is consistent with 
#the way that the R network plot package reads the adjacency matrix.
w.adj.scaled<-mat.1

#####################
#Export w.adj.scaled#
#####################
#Insert row and column labels in w.adj.scaled
row.names <- c(names.var)
col.names<-c("",names.var)
mat1<-cbind(row.names,w.adj.scaled)
mat2<-rbind(col.names,mat1)
write.table(mat2,"Adjacency.csv",col.names=FALSE,row.names=FALSE,sep=",")
