################################
#Nonlinear Time Series Analysis#
#AAFT surrogats                #
################################

#######################################################
#Contact: R. Huffaker                                 #   
#Professor                                            #
#Department of Agricultural and Biological Engineering#
#Institute of Food and Agricultural Sciences          #
#University of Florida                                #
#rhuffaker@ufl.edu                                    #
#######################################################

#####################################################################################
#Description:  This code reconstructs phase space for a time series using the       # 
#time-delay method. It calculates discriminating statistics for the reconstructed   #
#attractor series. It generates surrogates to test the null hypothesis that         #
#observed data are generated by a linear stochastic process (Theiler et al, 1992).  # 
#The null hypothesis is tested with a rank order test(Schreiber and Schmitz, 2000). #
#Three discriminating measures are tested: correlation dimension, Lyapunov exponent,#
#and nonlinear prediction skill.                                                    #
#####################################################################################

#####################################################
#Input:  Time series and parameters described below.#
#####################################################

##################################################################################
#Output: Printed statement at end of run stating whether Ho is accepted/rejected,#
#and exported table with test results.                                           #
##################################################################################

############
#References#
############
#Package 'fractal', Authors: William Constantine and Donald Percival (Applied Physics
  #Laboratory, University of Washington)
#Package 'tseriesChaos', Author: Antonio, Fabio Di Narzo
#Schreiber, T. and A. Schmitz. (2000) Surrogate times eries. Physica D 142:346-382.
#Small, M., and C. Tse. (2002) Applying the method of surrogate data to cyclic time 
#  series.  Physica D 164(2002):187-201.
#Theiler, J., S. Eubank, A. Longtin, B. Galdrikian, and J. Farmer. (1992)  Testing 
#  for nonlinearity in time series: The method of surrogate data.  Physica D 58:77-94.

rm(list=ls(all=TRUE))

#############################
#Load User-Defined Functions#
#############################
dump("embed_udf", file="embed_udf.R");source("embed_udf.R")
dump("mars_udf", file="mars_udf.R");source("mars_udf.R")
dump("mnse_udf", file="mnse_udf.R");source("mnse_udf.R")
dump("predict.is_udf", file="predict.is_udf.R");source("predict.is_udf.R")
dump("ds_udf", file="ds_udf.R");source("ds_udf.R")
dump("surr.cd_udf", file="surr.cd_udf.R");source("surr.cd_udf.R")
dump("surr.LE_udf", file="surr.LE_udf.R");source("surr.LE_udf.R")
dump("surr.np_udf", file="surr.np_udf.R");source("surr.np_udf.R")

###########
#Load Data#
###########
ts<-read.csv(".csv")
x<-ts$

################
#Set Parameters#
################
k=1
alpha=0.05
method="aaft"
r=0.20
fraction.fb=0.6
p=2

#k=determines number of surrogates

#alpha=probability of rejecting true null hypothesis that observed data are generated
  #by linear stochastic process

#method=Surrogates are generated with R-routine 'surrogate{fractal}' which gives a
  #choice of the following types: "aaft" (Theiler's Amplitude Adjusted Fourier Transform), 
  #"phase" (Theiler's phase randomization), "ce" (Davies and Harte's Circulant Embedding), 
  #or "dh" (Davison and Hinkley's phase and amplitude randomization). For example, 
  #"aaft surrogates test for linear-filtered noise, and are desinged to preserve both 
  #the power spectrum and probability distribution of the observed data. A Gaussian time 
  #series is generated that has the same length as the observed data and is reordered to 
  #have the same rank distribution. [Two data sets have the same rank distribution when 
  #the ith value in both sets are the jth largest for all i and j (Small & Tse, 2003)]. 
  #The Fourier transform is taken of the reordered Gaussian series, and the phases are 
  #randomized. A surrogate data set results from reordering the observed time series to 
  #have the same rank distribution as the inverse Fourier transform. 

#r = radius proportion. The radius proportion is used to compute the neighborhood of points 
  #considered to be 'nearest neighbors' in calculating the correlation dimension and maximum 
  #Lyapunov exponent. The nearest-neighbors neighborhood is computed as a proportion 'r' of 
  #the maximum distance between rows in the embedded time series matrix 'Mx' (i.e., between 
  #points on the reconstructed attractor). The proportion is set to avoid over-recruiting 
  #'nearest neighbors'. If the code crashes for "inability of false.nearest(tseriesChaos) to 
  #recruit nearest neighbors', then 'r' should be increased.

#fraction.fb  Fraction of data in forecasting base

#p = Nash-Sutcliffe parameter. If p = 1 (absolute deviations), p = 2(squared
     #deviations

##########################################################
#Step 1: Compute disciminating statistics for time series#
##########################################################
 results.ds_udf<-ds_udf(x,r,fraction.fb,p) #ds() user-defined
 cd.x<-results.ds_udf[[1]]  #correlation dimension for time series (m=2 or 3)
 cd.x.m<-results.ds_udf[[2]]  #correlation dimension for increasing embedding dimension
 LE.x<-results.ds_udf[[3]]  #max Lyapunov exponent for time series 
 LE.s<-results.ds_udf[[4]]   #Lyapunov exponents for increasing time steps
 m<-results.ds_udf[[5]] #embedding dimension, if m > 3, principal components used 
 d<-results.ds_udf[[6]]  #delay
 Mx<-results.ds_udf[[7]]  #embedded data for graphing phase space reconstruction
 np.x<-results.ds_udf[[8]]  #relative forecast error
 val.points<-results.ds_udf[[9]]  #denoised data in validation base
 pred.points<-results.ds_udf[[10]]  #predictions for comparison against validation base
   
#######################################################
#Step 2: Hypothesis testing with rank-order statistics#
#######################################################
#The null hypothesis Ho is that deterministic structure detected in the data is
#most likely due to a mimicking linear stochastic process. The alternative 
#hypothesis Ha is that deterministic structure is real. We run two-tailed tests
#for correlation dimension and Lyapunov exponent, and a lower tailed test
#for nonlinear prediction which tests whether attractor reconstructed from observed 
#data predicts with higher Nash-Sutcliffe efficiency than attractors reconstructed 
#from surrogate data.

 #######################
 #Correlation dimension#
 #######################
 results.surr_cd<-surr_cd_udf(x,method=method,k,alpha,one.sided=FALSE,r,slope.cd)
 cd.low<-results.surr_cd[1];cd.high<-results.surr_cd[2]

 ###################
 #Lyapunov Exponent#
 ###################
 results.surr_LE<-surr_LE_udf(x,method=method,k,alpha,one.sided=FALSE,r,slope.LE)
 LE.low<-results.surr_LE[1];LE.high<-results.surr_LE[2]

 ######################
 #Nonlinear Prediction#
 ######################
 results.surr_np<-surr_np_udf(x,method=method,k,alpha,one.sided=TRUE,r,fraction.fb,p)
 np.low<-results.surr_np[1];np.high<-results.surr_np[2]

########################
#Step 3: Generate plots#
########################
 setwd(dir.results)
 par(mfrow=c(2,2))
  
 ######################
 #Observed Time Series#
 ######################
 x.plot<-plot(x,type="l",main="Time Series",xlab="",ylab="")
 
 ############################
 #Phase space reconstruction#
 ############################
 if(m<=2)
  {PSR<-plot(Mx[,1],Mx[,2],xlab="x(t)",ylab="x(t-d)",main="Phase Space Reconstruction",
  type="l")} else 
  {library(scatterplot3d)
   if (m<=3)          
    {PSR<-scatterplot3d(Mx,main="Phase Space Reconstruction",xlab="x(t)",ylab="x(t-d)",
    zlab="x(t-2d)",box=TRUE,type="l",label.tick.marks=FALSE)} else 
    {v=cbind(Mx[,1],Mx[,2],Mx[,3]) #if m>3, do a 3D projection of Mx scatterplot   
     PSR<-scatterplot3d(v,main="Phase Space Reconstruction: 3D Projection",xlab="x(t)",
     ylab="x(t-d)",zlab="x(t-2d)",box=TRUE,type="l",label.tick.marks=FALSE)}}

   #Correlation dimension for increasing embedding dimensions
   em.dim<-1:length(cd.x.m)
 
      #Cancel default x-axis (xaxt="n") to define a new one with only integers 
      cd.m.plot<-plot(em.dim,cd.x.m,main="Correlation Dimension",xaxt="n",xlab="",
                      ylab="correlation dimension",pch=15,type="b")
      axis(side=1,at=1:length(cd.x.m)) #define custom x-axis, side=1 is x-axis, "at" 
                                       #defines 'tick-mark' numbers
      mtext(side=1,"embedding dimension",line=2,cex=0.8) #label custom x-axis, "line=2.5" puts
                                                      #label below tick-mark numbers
      #Label points in plot 
      cd.x.m<-round(cd.x.m,digits=2)
      text(em.dim,cd.x.m,cd.x.m,cex=0.7,pos=1,offset=0.4) 

dev.copy(png,"phase space reconstruction.png")
dev.off() 

#########
#Results#
#########
results1<-c(k,alpha,m,d,r)
#results1 is a column vector. Convert it to 3x5 matrix by filling in the 2-5th 
#columns with NULL. This allows it to be added to surrogate results below.
matrix.na<-matrix(rep("NULL",8),nrow=5,ncol=2)
results1b<-cbind(results1,matrix.na)

results2<-c(cd.x,cd.low,cd.high,LE.x,LE.low,LE.high,np.x,np.low,np.high)
results2a<- matrix(results2,nrow=3,ncol=3,byrow=TRUE)
results2b<-rbind(results2a,results1b)

row.names <- c("Correlation Dimension","Lyapunov Exponent","Predictive Skill","k",
               "Significance level","Embedding Dimension","Embedding Delay","r")
col.names<-c("","Time Series","Surr (low)","Surr (high)")
results2c<-cbind(row.names,results2b)
results2d<-rbind(col.names, results2c)
write.table(results2d,"nlts.results.csv",row.names=FALSE,col.names=FALSE,sep=",")

################################
#Error messages and corrections#
################################

#(1) If 'r' is set too low, so that the radius for finding 'nearest neighbors' is
     #too small, R may return the following error messages: "Error in lyap_k(x.em.vec, 
     #m = m.vec, d = d.vec, t = theiler(x), k = 1,  : not enough neighbours found" or
     #"Error: cannot allocate memory block of size 68719476736.0 Gb". A correction is 
     #to increase 'r'.  If 'r' is set too high, the correlation dimension may be 
     #underestimated.
 
#(2) If 'lb' and 'ub' are set too close together, so that the neighborhood for the 
     #random walk is too small, R may return the following error message: "Error in 
     #sample(int, 1) : invalid first argument". A correction is to increase the 
     #distance between 'lb' and 'ub'.

#(3) "Error: NA/NaN/Inf in foreign function call (arg 1)" This comes up if some 
     #columns in data matrix are longer than others. R wants them the same length, 
     #or it will fill in NAs in the shorter columns.

#(4) "Error in mars(ind, dep) : NA/NaN/Inf in foreign function call (arg 5)"  Increase
     #'r'.

#(5) If 'np' = NaN, increase fraction.fb

##########################################
#Hypothesis Result: Correlation Dimension#
##########################################
if(cd.x<=cd.low|cd.x>=cd.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}

##############################################
#Hypothesis Result: Maximum Lyapunov Exponent#
##############################################
if(LE.x<=LE.low|LE.x>=LE.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}

###############################################
#Hypothesis Result: Nonlinear Prediction Skill#
###############################################
if(np.x>=np.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}
