################################
#Nonlinear Time Series Analysis#
#PPS surrogates                #
################################

#######################################################
#Contact: R. Huffaker                                 #   
#Professor                                            #
#Department of Agricultural and Biological Engineering#
#Institute of Food and Agricultural Sciences          #
#University of Florida                                #
#rhuffaker@ufl.edu                                    #
#######################################################

####################################################################################
#Description:  This code reconstructs phase space for a time series using the      # 
#time-delay method. It calculates discriminating statistics for the reconstructed  #
#attractor series. It generates PPS surrogates to test the null hypothesis that    #
#observed data are generated by a linear stochastic process (Theiler et al, 1992). # 
#The null hypothesis is tested with a rank order test (Schreiber and Schmitz, 2000)#                                                                #
####################################################################################

#####################################################
#Input:  Time series and parameters described below.#
#####################################################

##################################################################################
#Output: Printed statement at end of run stating whether Ho is accepted/rejected,#
#and exported table with test results.                                           #
##################################################################################

############
#References#
############
#Package 'tseriesChaos', Author: Antonio, Fabio Di Narzo
#Schreiber, T. and A. Schmitz. (2000) Surrogate times eries. Physica D 142:346-382.
#Small, M., and C. Tse. (2002) Applying the method of surrogate data to cyclic time 
#  series.  Physica D 164(2002):187-201.
#Theiler, J., S. Eubank, A. Longtin, B. Galdrikian, and J. Farmer. (1992)  Testing 
#  for nonlinearity in time series: The method of surrogate data.  Physica D 58:77-94.

rm(list=ls(all=TRUE))

#############################
#Load User-Defined Functions#
#############################
dump("embed_udf", file="embed_udf.R");source("embed_udf.R")
dump("mars_udf", file="mars_udf.R");source("mars_udf.R")
dump("mnse_udf", file="mnse_udf.R");source("mnse_udf.R")
dump("predict.is_udf", file="predict.is_udf.R");source("predict.is_udf.R")
dump("ds_udf", file="ds_udf.R");source("ds_udf.R")
dump("PPS_adf", file="PPS_udf.R");source("PPS_udf.R")

###########
#Load Data#
###########
ts<-read.csv(".csv")
x<-ts$

################
#Set Parameters#
################
k=1
alpha=0.05
r=0.35
fraction.fb=0.6
lb=0.01;ub=0.4
p=2

#######################
#Variable descriptions#
#######################
#k=determines number of surrogates
#alpha=probability of rejecting true null hypothesis that observed data are generated
     #by linear stochastic process
#r = radius proportion. The radius proportion is used to compute the neighborhood 
     #of points considered to be 'nearest neighbors' in calculating the correlation 
     #dimension and maximum Lyapunov exponent. The nearest-neighbors neighborhood 
     #is computed as a proportion 'r' of the maximum distance between rows in the 
     #embedded time series matrix 'Mx' (i.e., between points on the reconstructed 
     #attractor). The proportion is set to avoid over-recruiting 'nearest neighbors'. 
     #If the code crashes for "inability of false.nearest(tseriesChaos) to recruit 
     #nearest neighbors', then 'r' should be increased.
#fraction.fb  Fraction of data in forecasting base
#p = Nash-Sutcliffe parameter. If p = 1 (absolute deviations), p = 2(squared
     #deviations

#############################
#Step 1: Generate surrogates#
#############################
nsurr.os=(k/alpha)-1   #one-sided hypothesis test 
nsurr.ts<-(2*k/alpha)-1  #two-sided hypothesis test
surr.os<-PPS_udf(x,nsurr.os,lb,ub) 
surr.ts<-PPS_udf(x,nsurr.ts,lb,ub) 

#########################################################
#Step 2:Compute disciminating statistics for time series#
#########################################################
   results.ds_udf<-ds_udf(x,r,fraction.fb,p) #ds() user-defined
   cd.x<-results.ds_udf[[1]]  #correlation dimension for time series (m=2 or 3)
   cd.x.m<-results.ds_udf[[2]]  #correlation dimension for increasing embedding dimension
   LE.x<-results.ds_udf[[3]]  #max Lyapunov exponent for time series 
   LE.s<-results.ds_udf[[4]]   #Lyapunov exponents for increasing time steps
   m<-results.ds_udf[[5]] #embedding dimension, if m > 3, principal components used 
   d<-results.ds_udf[[6]]  #delay
   Mx<-results.ds_udf[[7]]  #embedded data for graphing phase space reconstruction
   np.x<-results.ds_udf[[8]]  #relative forecast error
   val.points<-results.ds_udf[[9]]  #denoised data in validation base
   pred.points<-results.ds_udf[[10]]  #predictions for comparison against validation base

################################################ 
#Step 3: Discriminating measures for surrogates#
################################################ 
   cd.surr<-matrix(0,nsurr.os,1)  #correlation dimension
   LE.surr<-matrix(0,nsurr.os,1)  #Lyapunov exponent
   for(i in 1:nsurr.os){
      results.ds_udf<-ds_udf(surr.os[,i],r,fraction.fb,p)
      cd.surr[i,1]<-results.ds_udf[[1]]
      LE.surr[i,1]<-results.ds_udf[[3]]
      print("cd.surr,LE.surr")
      print(cbind(cd.surr,LE.surr))
   }  #end iloop
   
   LE.surr<-na.omit(LE.surr) #remove 'NA' elements from 'mxLE.surr' 
   
   np.surr<-matrix(0,nsurr.ts,1)  #nonlinear prediction
   for(j in 1:nsurr.ts){
      results.ds_udf<-ds_udf(surr.ts[,j],r,fraction.fb,p)
      np.surr[j,1]<-results.ds_udf[[8]]
      print("np.surr")
   }  #end jloop
   
   np.surr<-na.omit(np.surr)

###########################  
#Step 4:  Rank-order tests#
###########################
#Establish k smallest and largest values
   cd.low<-sort(cd.surr)[k];cd.high<-sort(cd.surr)[length(cd.surr)-(k-1)]
   LE.low<-sort(LE.surr)[k];LE.high<-sort(LE.surr)[length(LE.surr)-(k-1)]
   np.low<-sort(np.surr)[k];np.high<-sort(np.surr)[length(np.surr)-(k-1)]

#Step 6: Generate plots

   setwd(dir.results) 
   #Observed Time Series
   x.plot<-plot(x,type="l",main="Time Series",xlab="",ylab="")

   dev.copy(png,"time series plots.png")
   dev.off() 

   #Phase space reconstruction, complexity measures, surrogate histograms
   par(mfrow=c(3,2))

   #Phase Space Reconstruction
   if(m<=2){
      PSR<-plot(Mx[,1],Mx[,2],xlab="x(t)",ylab="x(t-d)",main="Phase Space Reconstruction",
             type="l")
      } else {
      library(scatterplot3d)
   if (m<=3){          
      PSR<-scatterplot3d(Mx,main="Phase Space Reconstruction",xlab="x(t)",ylab="x(t-d)",
                         zlab="x(t-2d)",box=TRUE,type="l",label.tick.marks=FALSE)
      } else {
      v=cbind(Mx[,1],Mx[,2],Mx[,3]) #if m>3, do a 3D projection of Mx scatterplot   
      PSR<-scatterplot3d(v,main="Phase Space Reconstruction: 3D Projection",xlab="x(t)",
                         ylab="x(t-d)",zlab="x(t-2d)",box=TRUE,type="l",label.tick.marks=FALSE)
      }
   }

#########
#Results#
#########
results1<-c(k,alpha,m,d,r)
#results1 is a column vector. Convert it to 3x5 matrix by filling in the 2-5th 
#columns with NULL. This allows it to be added to surrogate results below.
matrix.na<-matrix(rep("NULL",8),nrow=5,ncol=2)
results1b<-cbind(results1,matrix.na)

results2<-c(cd.x,cd.low,cd.high,LE.x,LE.low,LE.high,np.x,np.low,np.high)
results2a<- matrix(results2,nrow=3,ncol=3,byrow=TRUE)
results2b<-rbind(results2a,results1b)

row.names <- c("Correlation Dimension","Lyapunov Exponent","Predictive Skill","k",
               "Significance level","Embedding Dimension","Embedding Delay","r")
col.names<-c("","Time Series","Surr (low)","Surr (high)")
results2c<-cbind(row.names,results2b)
results2d<-rbind(col.names, results2c)
write.table(results2d,"nlts.results.csv",row.names=FALSE,col.names=FALSE,sep=",")

################################
#Error messages and corrections#
################################

#(1) If 'r' is set too low, so that the radius for finding 'nearest neighbors' is
     #too small, R may return the following error messages: "Error in lyap_k(x.em.vec, 
     #m = m.vec, d = d.vec, t = theiler(x), k = 1,  : not enough neighbours found" or
     #"Error: cannot allocate memory block of size 68719476736.0 Gb". A correction is 
     #to increase 'r'.  If 'r' is set too high, the correlation dimension may be 
     #underestimated.
 
#(2) If 'lb' and 'ub' are set too close together, so that the neighborhood for the 
     #random walk is too small, R may return the following error message: "Error in 
     #sample(int, 1) : invalid first argument". A correction is to increase the 
     #distance between 'lb' and 'ub'.

#(3) "Error: NA/NaN/Inf in foreign function call (arg 1)" This comes up if some 
     #columns in data matrix are longer than others. R wants them the same length, 
     #or it will fill in NAs in the shorter columns.

#(4) "Error in mars(ind, dep) : NA/NaN/Inf in foreign function call (arg 5)"  Increase
     #'r'.

#(5) If 'mnse' = NaN, increase fraction.fb

##########################################
#Hypothesis Result: Correlation Dimension#
##########################################
if(cd.x<=cd.low|cd.x>=cd.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}

##############################################
#Hypothesis Result: Maximum Lyapunov Exponent#
##############################################
if(LE.x<=LE.low|LE.x>=LE.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}

###############################################
#Hypothesis Result: Nonlinear Prediction Skill#
###############################################
if(np.x>=np.high)
    {print("Reject Ho: Data not generated by linear stochastic process")} else 
    {print("Accept Ho: Data generated by linear stochastic process")}
