###########################
#Phenomenological Modeling#
#Four Time Series         #
###########################

#######################################################
#Contact: R. Huffaker                                 #   
#Professor                                            #
#Department of Agricultural and Biological Engineering#
#Institute of Food and Agricultural Sciences          #
#University of Florida                                #
#rhuffaker@ufl.edu                                    #
#######################################################

#############
#Description#
#############
#This code empirically reconstructs an unknown dynamic system of ODEs from 
#four (observed) time series variables (x,y,z,w). Baker et al. solved this 
#problem for a known dynamic system fit to the four time series.  They 
#estimated the derivatives of x,y,z,w from their time series with centered 
#finite differences. Wei-Dong et al. used one time series (x) to reconstruct 
#an ODE system for the first, second and third time derivatives of x (so that 
#phase space is in coordinates of x and its first and second derivatives). 
#The derivatives of x are estimated with centered finite differences. Since, 
#they assumed that the ODE functions are unknown, they approximated them with
#polynomial expansions. Since, we assume that the ODEs for x,y,z,w are unknown, 
#we follow Wei-Dong and approximate them with polynomials expansions. The 
#methods proceeds as follows:

#(1) Compute 4th order centered differences of x,y,z,w 
#[Baker et al., Eq. 4]

#(2) Compute polynomial expansions for time derivatives of x,y,z,w 
#[Wei-Dong et al., p. 1368, Eq(6)]

#(3) Since ODEs are linear in the parameters, use Ordinary Least Squares 
#to fit differences to polynomial expansions. Apply Ritter and Munoz-Carpena
#(2013) for fit evaluation. 
 
#(4) Solve estimated ODE system for phase space dynamics (Soetaert et al.).

###################################
#Input: Time series and parameters#
###################################

############################################################################
#Outputs: Tables with coefficient estimates, simulated values, and goodness#
#of fit measures.                                                          #
############################################################################                                                            

############
#References#
############
#[1] Baker, G., Gollub, J. and Blackburn, J. (1996) Inverting chaos: Extracting
#    system parameters from experimental data. Chaos 6:528-533.
#[2] Wei-Dong, L., Ren, K., Meunier-Guttin-Cluzel, S. and Gouesbet, G. (2003)
#    Global vector-field reconstruction of nonlinear dynamical systems from a 
#    time series with SVD method and validation with Lyapunov exponents. Chinese 
#    Physics 12(12):1366-1373.
#[3] Gouesbet, G. and Maquet, J. (1992) Constructin of phenomenological models
#    from numerical time series. Physica D 58:202-215.
#[4] Mevik, B. and Wehrens, R. (2007) The pls package: Principal compenent and 
#    partial least squares regression in R. Journal of Statistical Software 
#    18(2):1-24.
#[5] Soetaert, K., Petzoldt, T. and Setzer, R. (Package deSolve: Solving initial
#    value differential equations in R.
#[6] Ritter, A. and Munoz-Carpena, R. "Performance evaluation of hydrologic models: 
#    statistical significance for reducing subjectivity in goodness-of-fit 
#    assessments," Journal Of Hydrology, vol. 480, pp. 33-45, 2013.
#[7] Package 'mpoly'.  Author: Kahle, David.
#[8] Package 'deSolve'. Authors: Soetaert, K., Petzoldt, T., and Setzer, R.
#[9] Package 'boot'. Authors: Cantry, A. and Ripley, B. 

rm(list=ls(all=TRUE))

################
#Load Libraries#
################
library(mpoly) #symbolic polynomial package
library(deSolve) #ODE solver package
library(boot)  #goodness of fit

#############################
#Load User Defined-Functions#
#############################
dump("mnse_udf", file="mnse_udf.R")
source("mnse_udf.R")
dump("maxima_udf", file="maxima_udf.R")
source("maxima_udf.R")

###########
#Load Data#
###########
ts<-read.csv(".csv")
x.ts<-ts$

################
#Set parameters#
################
#Polynomial expansion: Mp sets polynomial range. For example, if Mp=2, 
#then monomials (i.e., individual terms of polynomial) include x^0 
#(constants) and x^1 (single powers). Mp includes squared terms/interations.
Mp<-2
m<-4  #embedding dimension
Np<-Mp^m  #number of monomials in polynomial expansion

#The number of iterations in integration equals: T/delta.t, where T endtime 
#for in integration command, and delta_t is the time step. Example: 1/0.1 = 10 
#iterations, and 1/0.00001 = 100,000 iterations. The the integration error 
#(i.e., how much solved value deviates from observed value)is a power function 
#of the time step delta_t. In Euler, the error is roughly the delta_t^2.  For 
#stiff equations, a good estimate of delta_t is 0.00001 (documentation to 
#R: Implicit Runge-Kutta RADAU IIA).
delta_t<-1

###############################################
#Step 1: Compute 4th order centered difference#
###############################################

diff<-
  function(x,delta_t){  #formula for equally spaced grid
  #first centered observation is x.ts(3)
  x1<-x[3:(length(x)-2)] #first and last two observations lost
  x_plus1<-x[4:length(x)]
  x_plus2<-x[5:length(x)] 
  x_minus1<-x[2:length(x)]
  x_minus2<-x[1:length(x)]
  g<-length(x1);length(x_plus2)<-g;length(x_plus1)<-g
  length(x_minus1)<-g;length(x_minus2)<-g
  #Compute 4th order centered difference for equally spaced data points
  diff<-(-x_plus2+(8*x_plus1)-(8*x_minus1)+x_minus2)/(12*delta_t)
  results<-list(diff) 
  return(results)  
} #end user defined function

x.diff<-unlist(diff(x.ts,delta_t))  #finite difference for x.ts
y.diff<-unlist(diff(y.ts,delta_t))  #finite difference for y.ts
z.diff<-unlist(diff(z.ts,delta_t))  #finite difference for z.ts
w.diff<-unlist(diff(w.ts,delta_t))  #finite difference for w.ts

  ####################################################
  #Make time series compatible with finite difference#
  ####################################################
g<-length(x.diff)
#Each time series loses first and last two observations (4 total). 
#To make same length as *.diff, remove two observations from
#each end.
x<-x.ts[3:length(x.ts)]  #remove 1st two observations
y<-y.ts[3:length(y.ts)]
z<-z.ts[3:length(z.ts)]
w<-w.ts[3:length(w.ts)]
length(x)<-g;length(y)<-g  #remove last two observations
length(z)<-g;length(w)<-g 

###############################
#Step 2: Polynomial expansions#
###############################
#Set up polynomial expansion terms: Compute alpha's: powers of x,y,z,w in 
#polynomials [Wei-Dong et al., Eq(6)]

alpha1<-matrix(0,Np);alpha2<-matrix(0,Np);alpha3<-matrix(0,Np)
alpha4<-matrix(0,Np)

for(n in 1:Np){
alpha1a<-floor((n-1)/(Mp^(m-1)))
alpha2a<-floor(((n-1)-(alpha1a*(Mp^(m-1))))/(Mp^(m-2)))
alpha3a<-floor(((n-1)-(alpha1a*(Mp^(m-1)))-(alpha2a*(Mp^(m-2))))/(Mp^(m-3)))
alpha4a<-floor(((n-1)-(alpha1a*(Mp^(m-1)))-(alpha2a*(Mp^(m-2)))-
              (alpha3a*(Mp^(m-3))))/(Mp^(m-4)))
alpha1[n]<-alpha1a
alpha2[n]<-alpha2a
alpha3[n]<-alpha3a
alpha4[n]<-alpha4a
}

#'powers' matrix: Rows are powers of x (1st column), y (2nd column), and z 
#(3rd column) in monomials (mono)
powers<-cbind(alpha1,alpha2,alpha3,alpha4)

#Populate monomial terms with x,y,z,w data

termList2<-matrix(0,1) #store monomial terms in mpoly language
data<-matrix(0,length(x),Np) #store data for monomial terms
for(i in 1:Np){
  termList<-list(c(x=powers[i,1],y=powers[i,2],z=powers[i,3],
                   w=powers[i,4],coef=1))
  mono<-mpoly(termList)
  print(mono)
  mono.data<-x^powers[i,1]*y^powers[i,2]*z^powers[i,3]*w^powers[i,4]
  data[,i]<-mono.data
  termList2[i]<-mono  #store mono in mpoly language
  }
poly_data<-data  #data on monomial regressors
poly<-mpoly(termList2) #convert mpoly language to polynomial

#Insert column names into poly_data matrix (needed for regression)
if(Mp==2) {col.names=c("constant","w","z","zw","y","yw","yz","yzw","x","xw",
           "xz","xzw","xy","xyw","xyz","xyzw")}

if(Mp==3) {col.names=c("constant","w","w2","z","zw","zw2","z2","z2w","z2w2","y",
 "yw","yw2","yz","yzw","yzw2","yz2","yz2w","yz2w2","y2","y2w","y2w2","y2z","y2zw",
 "y2zw2","y2z2","y2z2w","y2z2w2","x","xw","xw2","xz","xzw","xzw2","xz2","xz2w",
 "xz2w2","xy","xyw","xyw2","xyz","xyzw","xyzw2","xyz2","xyz2w","xyz2w2","xy2",
 "xy2w","xy2w2","xy2z","xy2zw","xy2zw2","xy2z2","xy2z2w","xy2z2w2","x2","x2w",
 "x2w2","x2z","x2zw","x2zw2","x2z2","x2z2w","x2z2w2","x2y","x2yw","x2yw2","x2yz",
 "x2yzw","x2yzw2","x2yz2","x2yz2w","x2yz2w2","x2y2","x2y2w","x2y2w2","x2y2z",
 "x2y2zw","x2y2zw2","x2y2z2","x2y2z2w","x2y2z2w2")}

colnames(poly_data,do.NULL=FALSE)
colnames(poly_data)<-c(col.names)

####################
#Step 3: Regression#
####################
#Four dependent variables: x.diff,y.diff,z.diff,w.diff
#Data on exogenous variables (nxk)
exog.var1<-poly_data
n.obs<-nrow(exog.var1)  #Number of observations

#remove 1st column of 1's (regression package doesn't need to estimate constant)
exog.var<-exog.var1[,-1]
k<-ncol(exog.var)  #number of regressors excepting constant

#Construct [nx(k+1)] data frames needed by regression package
#1st column is dependent variable, subsequent k columns regressors
x.data1<-cbind(x.diff,exog.var);x.data<-as.data.frame(x.data1)
y.data1<-cbind(y.diff,exog.var);y.data<-as.data.frame(y.data1)
z.data1<-cbind(z.diff,exog.var);z.data<-as.data.frame(z.data1)
w.data1<-cbind(w.diff,exog.var);w.data<-as.data.frame(w.data1)

mono.terms<-c("const","w","z","zw","y","yw","yz","yzw","x","xw",
 "xz","xzw","xy","xyw","xyz","xyzw")
xeqn<-x.diff~w+z+zw+y+yw+yz+yzw+x+xw+xz+xzw+xy+xyw+xyz+xyzw
yeqn<-y.diff~w+z+zw+y+yw+yz+yzw+x+xw+xz+xzw+xy+xyw+xyz+xyzw
zeqn<-z.diff~w+z+zw+y+yw+yz+yzw+x+xw+xz+xzw+xy+xyw+xyz+xyzw
weqn<-w.diff~w+z+zw+y+yw+yz+yzw+x+xw+xz+xzw+xy+xyw+xyz+xyzw

results.x.ols<-lm(xeqn,data=x.data)  #run OLS
results.y.ols<-lm(yeqn,data=y.data)
results.z.ols<-lm(zeqn,data=z.data)
results.w.ols<-lm(weqn,data=w.data)

beta.x.ols<-as.numeric(coef(results.x.ols))  #OLS coefficients
beta.y.ols<-as.numeric(coef(results.y.ols))
beta.z.ols<-as.numeric(coef(results.z.ols))
beta.w.ols<-as.numeric(coef(results.w.ols))
xp<-beta.x.ols;yp<-beta.y.ols;zp<-beta.z.ols;wp<-beta.w.ols

#OLS diagnostics
summary(results.x.ols);summary(results.y.ols);summary(results.z.ols)
summary(results.w.ols)

##################################################
#Step 4: Estimated differentials from regressions#
##################################################
#Regression parameters
x1=xp[1];x2=xp[2];x3=xp[3];x4=xp[4];x5=xp[5];x6=xp[6];x7=xp[7];x8=xp[8];x9=xp[9]
  x10=xp[10];x11=xp[11];x12=xp[12];x13=xp[13];x14=xp[14];x15=xp[15];x16=xp[16]
y1=yp[1];y2=yp[2];y3=yp[3];y4=yp[4];y5=yp[5];y6=yp[6];y7=yp[7];y8=yp[8];y9=yp[9]
  y10=yp[10];y11=yp[11];y12=yp[12];y13=yp[13];y14=yp[14];y15=yp[15];y16=yp[16]
z1=zp[1];z2=zp[2];z3=zp[3];z4=zp[4];z5=zp[5];z6=zp[6];z7=zp[7];z8=zp[8];z9=zp[9]
  z10=zp[10];z11=zp[11];z12=zp[12];z13=zp[13];z14=zp[14];z15=zp[15];z16=zp[16]
w1=wp[1];w2=wp[2];w3=wp[3];w4=wp[4];w5=wp[5];w6=wp[6];w7=wp[7];w8=wp[8];w9=wp[9]
  w10=wp[10];w11=wp[11];w12=wp[12];w13=wp[13];w14=wp[14];w15=wp[15];w16=wp[16]

#Regression equations
dx<-x1+(x2*w)+(x3*z)+(x4*z*w)+(x5*y)+(x6*y*w)+(x7*y*z)+(x8*y*z*w)+(x9*x)+(x10*x*w)+
  (x11*x*z)+(x12*x*z*w)+(x13*x*y)+(x14*x*y*w)+(x15*x*y*z)+(x16*x*y*z*w)
dy<-y1+(y2*w)+(y3*z)+(y4*z*w)+(y5*y)+(y6*y*w)+(y7*y*z)+(y8*y*z*w)+(y9*x)+(y10*x*w)+
  (y11*x*z)+(y12*x*z*w)+(y13*x*y)+(y14*x*y*w)+(y15*x*y*z)+(y16*x*y*z*w)
dz<-z1+(z2*w)+(z3*z)+(z4*z*w)+(z5*y)+(z6*y*w)+(z7*y*z)+(z8*y*z*w)+(z9*x)+(z10*x*w)+
  (z11*x*z)+(z12*x*z*w)+(z13*x*y)+(z14*x*y*w)+(z15*x*y*z)+(z16*x*y*z*w)
dw<-w1+(w2*w)+(w3*z)+(w4*z*w)+(w5*y)+(w6*y*w)+(w7*y*z)+(w8*y*z*w)+(w9*x)+(w10*x*w)+
  (w11*x*z)+(w12*x*z*w)+(w13*x*y)+(w14*x*y*w)+(w15*x*y*z)+(w16*x*y*z*w)

##########################
#Step 5: Solve ODE system#
##########################
#Choice of integrators in deSolve: lsoda, lsode, lsodes, lsodar, 
#vode, daspk, euler, rk4, ode23, ode45, radau, bdf, 
#bdf_d, adams, impAdams, impAdams_d, iteration

#initial conditions set to initial observed data points
state<-c(x=x[1],y=y[1],z=z[1],w=w[1])

parameters<-c(x1=xp[1],x2=xp[2],x3=xp[3],x4=xp[4],x5=xp[5],x6=xp[6],x7=xp[7],
  x8=xp[8],x9=xp[9],x10=xp[10],x11=xp[11],x12=xp[12],x13=xp[13],x14=xp[14],
  x15=xp[15],x16=xp[16],y1=yp[1],y2=yp[2],y3=yp[3],y4=yp[4],y5=yp[5],y6=yp[6],
  y7=yp[7],y8=yp[8],y9=yp[9],y10=yp[10],y11=yp[11],y12=yp[12],y13=yp[13],
  y14=yp[14],y15=yp[15],y16=yp[16],z1=zp[1],z2=zp[2],z3=zp[3],z4=zp[4],z5=zp[5],
  z6=zp[6],z7=zp[7],z8=zp[8],z9=zp[9],z10=zp[10],z11=zp[11],z12=zp[12],z13=zp[13],
  z14=zp[14],z15=zp[15],z16=zp[16],w1=wp[1],w2=wp[2],w3=wp[3],w4=wp[4],w5=wp[5],
  w6=wp[6],w7=wp[7],w8=wp[8],w9=wp[9],w10=wp[10],w11=wp[11],w12=wp[12],w13=wp[13],
  w14=wp[14],w15=wp[15],w16=wp[16])

model<-function(t,state,parameters){
  with(as.list(c(state,parameters)),{
dx<-x1+(x2*w)+(x3*z)+(x4*z*w)+(x5*y)+(x6*y*w)+(x7*y*z)+(x8*y*z*w)+(x9*x)+(x10*x*w)+
  (x11*x*z)+(x12*x*z*w)+(x13*x*y)+(x14*x*y*w)+(x15*x*y*z)+(x16*x*y*z*w)
dy<-y1+(y2*w)+(y3*z)+(y4*z*w)+(y5*y)+(y6*y*w)+(y7*y*z)+(y8*y*z*w)+(y9*x)+(y10*x*w)+
  (y11*x*z)+(y12*x*z*w)+(y13*x*y)+(y14*x*y*w)+(y15*x*y*z)+(y16*x*y*z*w)
dz<-z1+(z2*w)+(z3*z)+(z4*z*w)+(z5*y)+(z6*y*w)+(z7*y*z)+(z8*y*z*w)+(z9*x)+(z10*x*w)+
  (z11*x*z)+(z12*x*z*w)+(z13*x*y)+(z14*x*y*w)+(z15*x*y*z)+(z16*x*y*z*w)
dw<-w1+(w2*w)+(w3*z)+(w4*z*w)+(w5*y)+(w6*y*w)+(w7*y*z)+(w8*y*z*w)+(w9*x)+(w10*x*w)+
  (w11*x*z)+(w12*x*z*w)+(w13*x*y)+(w14*x*y*w)+(w15*x*y*z)+(w16*x*y*z*w)
 list(c(dx,dy,dz,dw))
 }) #end with (as.list...
}  #end model function

times<-seq(0,length(x),by=delta_t)  #specify integration step
out<-ode(y=state,times=times,func=model,parms=parameters,method="rk4")
head(out)

#Siimulated variables
x.sim<-out[,"x"];length(x.sim)<-length(x)
y.sim<-out[,"y"];length(y.sim)<-length(y)
z.sim<-out[,"z"];length(z.sim)<-length(z)
w.sim<-out[,"w"];length(w.sim)<-length(w)

##########################################################
#Step 6: Goodness of fit (Ritter and Munoz-Carpena, 2013)#
##########################################################
#Reset x,y,z to vector values (deSolve sets x,y,z,w to their initial values
x<-x.ts[3:length(x.ts)];length(x)<-g
y<-y.ts[3:length(y.ts)];length(y)<-g
z<-z.ts[3:length(z.ts)];length(z)<-g
w<-w.ts[3:length(w.ts)];length(w)<-g

sim<-cbind(x.diff,y.diff,z.diff,w.diff)
signal<-cbind(dx,dy,dz,dw)
names<-colnames(sim)<-c("x.ode","y.ode","z.ode","w.ode")

#Define bootstrapping statistic: The R bootstraping routine works off a 
#user-defined function that calculates a statistic to be measured in each sample.

nse_udf_boot<-function(data,indices){ 
   d <- data[indices,]
   z1<-d[,1];z2<-d[,2]
   num<-sum((z1-z2)^2)
   den<-sum((z1-mean(z1))^2)
   nse<-1-(num/den)
   return(nse)   
} #end user-defined function

mnse_udf_boot<-function(data,indices){ 
   d <- data[indices,]
   z1<-d[,1];z2<-d[,2]
   num<-sum(abs(z1-z2))
   den<-sum(abs(z1-mean(z1)))
   mnse<-1-(num/den)
   return(mnse)   
} #end user-defined function

  ################
  #Set Parameters#
  ################
#For Nash-Sutcliffe with squared deviations
p=2
model=nse_udf_boot #see "b=boot()" below 
#For Nash-Sutcliffe with absolute deviations 
#p=1
#model<-mnse_udf_boot

#Loop i: interates through each time series and corresponding ssa 
#reconstruction. 

#Output
#hold.i: Nash-Sutcliffe efficiency for each time series
#hold.ii[,i]: 95% bootstrapped confidence interval on mnse
#hold.iii[,i]: P-value that Nash-Sutcliffe is < 0.65
#hold.iv: 11 rows equal to fitness probability categories calculated

hold.i<-matrix(0,1,length(names))
hold.ii<-matrix(0,2,length(names))
hold.iii<-matrix(0,1,length(names))
hold.iv<-matrix(0,11,length(names))
   for(i in 1:length(names)){ 
      w1.sim<-sim[,i]
      w1.sig<-signal[,i]
      data1<-cbind(w1.sim,w1.sig)

      #Calculate Nash-Sutcliffe Efficiency for each time series
      setwd(dir.udf)
      mnse.ssa<-mnse_udf(w1.sim,w1.sig,p)

      #Bootstrap Nash-Sutcliffe statistic
      b = boot(data=data1,statistic=model,R=5000)
      #b$t gives nse statistic for bootsraps (Rx1)
      #bca adjusts for bias and skewness in bootstrap distribution (Ritter and 
      #Munoz-Carpena, p.37)
      ci.out<-boot.ci(b, conf=0.95,type="bca") 
      ci<-ci.out$bca[4:5]

      #cumulative distribution function
      cdf<-ecdf(b$t)
      plot_cdf<-plot(cdf,main=names[i],ylab="cdf",xlab="mnse")
      dev.copy(png,filename=paste("cdf_plot_",i,".png",sep=""))
      dev.off()

      #compute cdf values
      tt <- seq(0,1, by = 0.001)
      cdf.val<-cdf(tt);#cdf.val<-unlist(cdf.val)

      #Probabilities of nse fit intervals
      Prob_less_0<-cdf(0)  #Prob(mnse<0) 
      Prob_0_0.9<-cdf(0.099)-cdf(0)       #Prob(0.0<mnse<0.099) 
      Prob_0.1_0.19<-cdf(0.199)-cdf(0.1)  #Prob(0.1<mnse<0.199)
      Prob_0.2_0.29<-cdf(0.299)-cdf(0.2)  #Prob(0.2<mnse<0.29) 
      Prob_0.3_0.39<-cdf(0.399)-cdf(0.3)  #Prob(0.3<mnse<0.399)
      Prob_0.4_0.49<-cdf(0.499)-cdf(0.4)  #Prob(0.4<mnse<0.49) 
      Prob_0.5_0.59<-cdf(0.599)-cdf(0.5)  #Prob(0.5<mnse<0.599) 
      Prob_0.6_0.69<-cdf(0.699)-cdf(0.6)  #Prob(0.6<mnse<0.69) 
      Prob_0.7_0.79<-cdf(0.799)-cdf(0.7)  #Prob(0.7<mnse<0.799) 
      Prob_0.8_0.89<-cdf(0.899)-cdf(0.8)  #Prob(0.8<mnse<0.899)
      Prob_0.9_1<-cdf(1)-cdf(0.9)         #Prob(0.9<mnse<1)

      fit.int1<-cbind(Prob_less_0,Prob_0_0.9,Prob_0.1_0.19,Prob_0.2_0.29,Prob_0.3_0.39,
         Prob_0.4_0.49,Prob_0.5_0.59,Prob_0.6_0.69,Prob_0.7_0.79,Prob_0.8_0.89,Prob_0.9_1)

      ####################
      #Hypothesis testing#
      ####################
      #H0: nse < nse0 = 0.65
      #HA: nse >= nse0 = 0.65

      #Compute probability in favor of H0: Prob(nse < 0.65) (Ritter and
      #Munoz-Carpena, pp. 38-39)
      pval<-cdf(0.65)

   hold.i[,i]<-mnse.ssa
   hold.ii[,i]<-ci
   hold.iii[,i]<-pval
   hold.iv[,i]<-fit.int1
   
}  #end loop i

#1st row (Nash-Sutcliffe for each time series)
#2nd and 3rd rows are lower and upper 95% confidence bound
results1<-rbind(hold.i,hold.ii,hold.iii,hold.iv) 

  ###########################################
  #Export probabilities of nse fit intervals#
  ###########################################
results2<- matrix(results1,nrow=15,ncol=length(names),byrow=FALSE,dimnames=
     list(c("mnse","95% low","95% high","P-value","Prob(mnse<0)","Prob(0<mnse<0.099)",
     "Prob(0.1<mnse<0.199)","Prob(0.2<mnse<0.299)","Prob(0.3<mnse<0.399)",
     "Prob(0.4<mnse<0.499)","Prob(0.5<mnse<0.599)","Prob(0.6<mnse<0.699)",
     "Prob(0.7<mnse<0.799)","Prob(0.8<mnse<0.899)","Prob(0.9<mnse<1)"),
     c(names[1:length(names)])))

###############
#Export Output#
###############
setwd(dir.results)
#Table with coefficients
coefs<-data.frame(mono.terms,beta.x.ols,beta.y.ols,beta.z.ols,beta.w.ols)
colnames(coefs)<-c("terms","x.diff","y.diff","z.diff","w.diff")
write.table(coefs,sep=",","coefs.csv",row.names=FALSE)

#Simulated values
sim.val<-cbind(x.sim,y.sim,z.sim,w.sim)
write.table(sim.val,sep=",",col.names=c("x.sim","y.sim","z.sim","w.sim"),
  "sim.val.csv",row.names=FALSE)

#Goodness of fit
write.csv(results2,"fit.int.csv")