#USER-DEFINED FUNCTION: surr_PPS(x=time series, nsurr=number of surrogate series',lb=lower 
                                 #neighborhood bound, ub=upper neighborhood bound)

#Output: Generates matrix whose columns are PPS Surrogates following the procedure 
#outlined in Small, M., and Tse, C. (2003). 

#Background: The PPS method implements a random walk over the attractor (i.e., across 
#the rows of the embedded time series matrix 'Mx') to test for the presence of a 'noisy 
#limit cycle'.  A starting point 'v1' is randomly selected on the attractor, and for a given 
#neighborhood around 'v1', another point 'v2' is randomly selected on the attractor, and so 
#on. This is done lts times, where lts is the length of the observed time series. The 
#PPS Surrogates are the first coordinates of the randomly-selected points. We implement the 
#random walk with the following procedure:

#1. Index the attractor points (i.e., the rows of embedding matrix x.em) in chronological 
#order:  Index     Point on attractor 
#         [1]          x1,y1,z1
#         [2]          x2,y2,z2  . . .

#2. Initialize the random walk by using sample(base) function to select an index [] 
#at random. 

#3. Establish the set of neighboring indices from which to draw the next index in the 
#random walk.

#a. First, determine the range of attractor points to be included in the neighborhood 
#of 'v1' by computing the distances from 'v1' to all of the remaining attractor points, 
#and scaling each distance so that it is a percentage of the maximum distance computed. 
#This differs from using the distance matrix to calculate the radius parameter required 
#to calculate correlation dimension and maximum Lyapunov exponent where the maximum 
#distance was taken from the entire matrix. In the random walk on the attractor, we use 
#take the maximum distance from a single column of the distance matrix containing 
#distances from a single point (e.g., initially 'v1') to other attractor points. 

#b. Second, identify the attractor points that rest above a given  lower bound 
#percentage of maximum distance and below a given upper bound.   Use the which(base) 
#function to identify the indices of these neighboring points.  [The which(base) 
#command identifies the indices of vector elements satisfying a given condition.  For 
#example: x = c(4,3,9,1), y = which(x<4) = 2,4 (i.e., the elements 3 and 1 satisfy the 
#condition x<4, and they are the 2nd and 4th elements in x, respectively. ]  This 
#determines the interval of indices from which to randomly select the next index.

#4. Repeat until there are a number of randomly-selected indices equaling the length 
#of the observed time series.  SPP Surrogates are composed of the first coordinates of 
#the randomly-selected points on the attractor (i.e., x1, x2, x3).  These coordinates 
#are in the first column Mx.c1 of the embedding matrix x.em.   To identify the point 
#on the attractor associated with a given randomly-selected index [k], evaluate Mx.c1[k].

#Parameters:

   #lb (0<lb<1) = lower neighborhood bound. The lowest fraction of maximum distance that 
                  #a point can have to be included in the neighborhood for the random walk.  

   #ub (0<lb<1) = upper neighborhood bound. The largest fraction of maximum distance that 
                  #a point can have to be included in the neighborhood for the random walk.

#Output: surrPPS = matrix of P


PPS_udf<-function(x,nsurr,lb,ub){

#Step 1: Embed time series
 library(tseriesChaos)
   Mx<-embedd(x,3,2)

   #Step 1a: Isolate first column of embedding matrix 'x.em' since PPS algorithm 
             #requires it later

      Mx.c1<-Mx[,1]  #isolate first column of embedding matrix 'Mx'

#Step 2: Establish neighborhood of points for random walk

   #Step 2a: Calculate distance matrix from embedded time series matrix 'Mx'
   
   library(fields)
   dist<-rdist(Mx) #rdist(fields) [euclidean distances in embedded time series matrix]
   
   #Step 2b: Scale each column element to maximum distance within that column in 'dist'

   sf<-apply(dist,2,max) #apply(base) [find maximum distance in each column]
   dist.tr<-t(dist)
   dist.sc<-t(dist.tr/sf)  #w/out transposing, division divides rows by max distince in
                           #each column

#Step 5: Random walk on attractor (i.e., over rows of embedded time series matrix 'x.em'.

   rt<-sample(length(Mx.c1),1)  #Initialize with randomly-selected time 'rt' on attractor     

   jloop<-matrix(0,length(x),nsurr)  #matrix to hold PPS surrogates as they are generated
                                     #by inner loop  
   for(j in 1:nsurr){     #outer loop creates multiple PPS surrogates
      rt.out<-matrix(0,length(x),1) #matrix to hold randomly-selected times in inner i loop
      iloop<-matrix(0,length(x),1)  #matrix to hold single PPS surrogate created in i loop 

      for(i in 1:length(x)){  #inner loop creates one PPS surrogate--each loop creates
                              #one element
         #index is element's position in vector, e.g., in v=(10,2) the index of v[1]=10 is 1,
         #index also represents time
         ngh.lb<-which(lb < dist.sc[,rt[i]]) #indices satisfying lower bound 'lb' of nghbhd 
         ngh.ub<-which(ub > dist.sc[,rt[i]]) #indices satisfying upper bound 'ub' of nghbhd
         int<-intersect(ngh.lb,ngh.ub)       #indices satisfying both bounds in neighborhood
         #randomly-select time next period from neighborhood of indices--if there is only one 
         #member of intersection use it as next time period
         if (length(int)==1) {rt[i+1]<-int} else {rt[i+1]<-sample(int,1)}
         rt.out[i,1]<-rt[i]  #rt has n=length(x)+1 elements (due to rt[i+1], need length(x))
         iloop[i,1]<-Mx.c1[rt.out[i]] #evaluate Mx.c1 (1st column x.em) at random times
      }   #end iloop     
  
      out1<-iloop   #PPS surrogate

      jloop[,j]<-out1

   }  #end jloop

   surr.PPS<-jloop 

   return(surr.PPS)
   
}  #end user-defined function PPS_udf()
